/* sd_window by Shadowdarke (shadowdarke@hotmail.com) 2006
sd_window provides special popup dialog windows with several controls that normal BYOND
browse() windows lack. sd_window lets you control the position of the window on screen and
triggers a proc when the window is closed.

The windows are controlled by the /sd_window datum. In its simplest form, you can display a
window with a single line of code just by creating the window with the content and target
arguments.

See the accompanying sd_window_demo.dm file for code examples.
*/

//
#define SDW_FOCUS		0x0001
#define SDW_RESIZE		0x0002
#define SDW_STATUS		0x0004
#define SDW_SCROLL		0x0008
#define SDW_SUNKEN		0x0010
#define SDW_HELP		0x0020

sd_window
	var
		x = -1				/* Horizontal position of the left side of window in pixels.
								Negative values center the window horizontally. Values larger
								than screen_width - window_width butt the window against the
								right side of the screen */
		y = -1				/* Vertical position of the left side of window in pixels.
								Negative values center the window vertically. Values larger
								than screen_height - window_height butt the window against the
								bottom of the screen */
		width = 300			// Width of the window
		height = 300		// Height of the window
		content				// The content displayed in the window
		timer = 0			/* Optional close timer. If set, the window will automatically close
								timer ticks after it opens. */
		flags	= SDW_RESIZE|SDW_SCROLL	/* Display flags for the window. Possible values include:
								SDW_FOCUS	If not set, window restores focus to Dream Seeker
								SDW_RESIZE	Window may be resized
								SDW_STATUS	Display the status bar
								SDW_SCROLL	Display scroll bars
								SDW_SUNKEN	Show a sunken window edge
								SDW_HELP	Display context sensitive help icon */
		list/resources		/* Resource files that the window display depends on. Each resource
								will be sent to any client when they view the window. If any
								resources are used, this will be an associated list:
									resources[filename] = BYOND_resource.
							 	Use the sd_window.AddResource() to add resources. */
		tmp
			viewers			// list of clients viewing this dialog
			resource_name	// filename used for client-side content file
			close_flag = 1	// so unenhanced window won't call Close() a second time

/* PROCS:
	New(content, target, timer, x, y, width, height, flags)
		Create an sd_window and optionally set its vars.
		ARGS:	all arguments are optional.
			target	- If set, the window will automatically display to target. See the Display()
						proc for more information.
			All other args set the sd_window var of the same name.
		RETURNS:	A reference to the created window

	AddResource(resource, filename)
		Adds resource to the window resources list. Use this proc to add resources that
			should be sent to the client before the window is displayed.
		ARGS:
			resource	- the BYOND resource
			filename	- the filename used to store the resource in the client's cache

	* AddViewer(target)
		Add the client(s) of target to the viewer's list. Display() calls this
			automatically.
		* This proc is used internally. You should not call or override it. Call Display()
			instead.

	Close(target, report = 1, remove = 1)
		Closes the window for all clients in target.
		ARGS:
			target	- mob, client, or a list of mobs and or clients
			report	- if set, call the window's Closed() proc
			remove	- if set, remove target from the window's viewers list.

	Closed(target)
		Called when the window is closed by target.
			Prototype function: override to do something when the window closes.
		ARGS
			target	- a mob, client, or list of mobs and/or clients that closed the
						window

	Display(target)
		Display the window to target
		ARGS
			target	- a mob, client, or list of mobs and/or clients to display the window to

	NoViewers()
		Called when no clients are viewing the window.
			Prototype function: override to do something when the window closes.
			By default, this deletes the window to free system resources.

	* RemoveViewer(target, track = 1)
		Remove target from the viewers list.
		* This proc is used internally. You should not call or override it. Call Close() instead.
		ARGS:
			target	- a mob, client, or list of mobs and/or clients
			track	- if set, remove this window from each target client's sd_windows list

	SendResources(target)
		Send content and all associated resources to target.
		ARGS:
			target	- a mob, client, or list of mobs and/or clients

	* Viewers(list/L)
		Update viewers list based on the value of L
		* This proc is used internally. You should not call or override it.
*/

	Del()
		if(viewers)
			Close(viewers)
		..()

	New(content, target, timer, x, y, width, height, flags)
		..()
		// if any of these are null (not specified), keep the default values
		if(!isnull(content)) src.content = content
		if(!isnull(x)) src.x = x
		if(!isnull(y)) src.y = y
		if(!isnull(width)) src.width = width
		if(!isnull(height)) src.height = height
		if(!isnull(flags)) src.flags = flags
		if(!isnull(timer)) src.timer = timer
		if(target) Display(target)

		resource_name = "~"
		// add a random 7 character string (base62 yields 3,521,614,606,208 unique strings)
		for(var/loop = 1 to 7)
			var/pick = rand(61)
			if(pick>35)
				pick = ascii2text(pick+61)
			else if(pick>9)
				pick = ascii2text(pick+55)
			else pick = "[pick]"
			resource_name += pick

	Topic(href, hlist[])
		if("CLOSED" in hlist)
			var/client/C = usr.client
			if(C && close_flag) Close(C)
		..()

	proc
		AddResource(resource, filename)
			/* Adds resource to the window resources list. Use this proc to add resources that
				should be sent to the client before the window is displayed.
				ARGS:
					resource	- the BYOND resource
					filename	- the filename used to store the resource in the client's cache */
			if(!resource || !filename)
				CRASH("Missing resource or filename.")
			if(!resources) resources = list()
			resources[filename] = resource

		AddViewer(target)
			/* Add the client(s) of target to the viewer's list. Display() calls this
				automatically. */
			var/list
				In = target
				Out = viewers
			if(!istype(In))
				if(target) In = list(target)
				else In = list()
			if(!istype(Out))
				if(viewers) Out = list(viewers)
				else Out = list()

			var/client/C
			for(var/I in In)
				if(istype(I, /mob))
					var/mob/M = I
					C = M.client
				else C = I
				if(istype(C) && !(C in Out))
					Out += C
					C.sd_AddWindow(src)

			Viewers(Out)

		Close(target, report = 1, remove = 1)
			/* Closes the window for all clients in target.
				ARGS:
					target	- mob, client, or a list of mobs and or clients
					report	- if set, call the window's Closed() proc
					remove	- if set, remove target from the window's viewers list. */
			close_flag = 0
			target << browse(null, "window=\ref[src]")
			if(report) Closed(target)
			if(remove) RemoveViewer(target)
			spawn(2) close_flag = 1

		Closed(target)
			/* Called when the window is closed by target.
				Prototype function: override to do something when the window closes.
				ARGS
					target	- a mob, client, or list of mobs and/or clients that closed the
								window */

		Display(target)
			/* Display the window to target
				ARGS
					target	- a mob, client, or list of mobs and/or clients to display the window
								 to */
			SendResources(target)
			AddViewer(target)


			var/list/In = target
			if(!istype(In))
				if(target) In = list(target)
				else In = list()

			var/client/C
			for(var/I in In)
				if(istype(I, /mob))
					var/mob/M = I
					C = M.client
				else if(istype(I, /client)) C = I
				else continue

				var/html
				if(C.sd_enhanced_windows)
					// display hidden controller window
					html = {"<script>\
						var N = null;\
						window.onload= function(){\
							N=window.showModelessDialog("[resource_name].html", window, \
								"[(x<0)?"":"dialogLeft:[x]px;"][(y<0)?"":"dialogTop:[y]px;"]\
								dialogHeight:[height]px;dialogWidth:[width]px;\
								[flags&SDW_SUNKEN?"edge:sunken;":""]\
								[flags&SDW_HELP?"":"help:0;"][flags&SDW_RESIZE?"resizable:1;":""]\
								[flags&SDW_SCROLL?"":"scroll:0;"]\
								[flags&SDW_STATUS?"":"status:0;"]"); \
							N.[flags&SDW_FOCUS?"focus":"blur"](); \
							[flags&SDW_FOCUS?"":"window.focus(); "]\
							[timer?"setTimeout(\"N.close()\",[timer*100]);":""]\
							setTimeout("tst()",100)\
						};\
						window.onunload = function(){\
							if(N) {\
								if(!N.closed) {\
									N.close(); \
								}\
							}\
						};\
						function tst() {\
							if(!N) {\
								document.location.href = 'BYOND://?src=\ref[src]&CLOSED'\
							} \
							else if(N.closed) {\
								document.location.href = 'BYOND://?src=\ref[src]&CLOSED'\
							} else { \
								setTimeout("tst()",100)\
							}\
						}\
						</script>"}
					C << browse(html, "window=\ref[src];size=0x0;can_resize=0")

				else	// don't use modeless windows
					var/title
					var/pos = findtext(content, "<title>")
					if(pos)
						title = copytext(content, pos+7, findtext(content, "</title>"))
					html = {"<head><title>[title]</title>
						<style>BODY {margin:0;}</style>\
						<script>document.oncontextmenu = function(){return false};\
						window.onunload = function(){\
							document.location.href = 'BYOND://?src=\ref[src]&CLOSED'\
						};\
						</script></head>\
						<body scroll=no>\
						<iframe src=[resource_name].html id=Frame width=100% height=100% \
						frameborder=no scrolling=[flags&SDW_SCROLL?"auto":"no"]>[content]</iframe></body>"}
					C << browse(html, "window=\ref[src];size=[width]x[height];\
						[flags&SDW_RESIZE?"":"can_resize=0;can_minimize=0"]")
					if(timer) spawn(timer) C << browse(null, "window=\ref[src]")


			if(!(flags&SDW_FOCUS))
				spawn(2) target << browse(null, null)	// short delay required to reset focus

		NoViewers()
			/* Called when no clients are viewing the window.
				Prototype function: override to do something when the window closes.
				By default, delete the window to free system resources. */
			del(src)

		RemoveViewer(target, track = 1)
			/* Remove target from the viewers list.
				ARGS:
					target	- a mob, client, or list of mobs and/or clients
					track	- if set, remove this window from each target client's sd_windows
								list*/
			if(!viewers) return
			var/list
				In = target
				Out = viewers
			if(!istype(In))
				if(target) In = list(target)
				else In = list()
			if(!istype(Out))
				if(viewers) Out = list(viewers)
				else Out = list()

			var/client/C
			for(var/I in In)
				if(istype(I, /mob))
					var/mob/M = I
					C = M.client
				else C = I
				if(istype(C))
					Out -= C
					if(track)
						C.sd_windows -= src
						if(!C.sd_windows.len) del(C.sd_windows)	// free the empty list

			Viewers(Out)

		SendResources(target)
			/* Send content and all associated resources to target.
				ARGS:
					target	- a mob, client, or list of mobs and/or clients */
			var/filename = "[resource_name].html"
			fdel(filename)
			text2file(content,filename)
			var/resource = fcopy_rsc(filename)
			fdel(filename)
			target << browse_rsc(resource, filename)

			if(resources)
				for(filename in resources)
					target << browse_rsc(resources[filename],filename)

		Viewers(list/L)
			// Update viewers list based on the value of L
			switch(L.len)
				if(0)
					viewers=null
					NoViewers()
				if(1)	viewers = L[1]
				else	viewers = L


sd_window/detect_navigator	// checks the OS and navigator of the target
	var/client/owner

	Del()
		if(owner)
			owner << browse(null,"window=\ref[src]")
			owner.sd_windows -= src
			if(!owner.sd_windows.len) del(owner.sd_windows)	// free the empty list
		..()


	New(C)
		..()
		owner = C
		if(!istype(owner)) del(src)
		var/html = {"<BODY onLoad=\
			'document.location.href="BYOND://?src=\ref[src]\
			&sd_os="+navigator.platform+"\
			&sd_brwsr="+navigator.appName+"\
			&sd_brver="+encodeURIComponent(navigator.appVersion)+"\
			&sd_brmin="+encodeURIComponent(navigator.appMinorVersion);'></BODY>"}
		owner << browse(html, "window=\ref[src];size=0x0;can_resize=0")
		owner.sd_AddWindow(src)

	Topic(href,hlist[])
		//world.log << href
		owner.sd_OS = hlist["sd_os"]
		owner.sd_browser_name = hlist["sd_brwsr"]
		owner.sd_browser_version = hlist["sd_brver"]
		owner.sd_browser_minor_version = hlist["sd_brmin"]
		del(src)


client
	var/tmp
		list/sd_windows				// windows that the client is viewing
		sd_enhanced_windows			// if set, this client may use modeless windows
		sd_OS						// operating system
		sd_browser_name				// name of the browser
		sd_browser_version			// version of the browser
		sd_browser_minor_version	// minor version of the browser

	Del()
		// remove from window viewers list so the window can free itself if not viewed by anyone
		for(var/sd_window/S in sd_windows)
			if((src == S.viewers) || (src in S.viewers))	S.RemoveViewer(src, 0)
		..()

	proc/sd_AddWindow(sd_window/S)
		// add S to the client's sd_windows list
		if(!sd_windows) sd_windows=list()
		if(!(S in sd_windows)) sd_windows += S
